!=========================================
module act_beam_chisq

! Module to calculated the correction to the ACT chisq from beam corrections
! Adapted from Mike Nolta's code for WMAP
!=========================================
  implicit none

  public :: init_act_beam_chisq
  public :: compute_act_beam_chisq
  public :: get_free_lun	

  private
  integer, parameter :: n_beam_modes = 10 
  integer :: nmodes
  integer :: bmax0
  real(kind=8), allocatable :: mode(:,:), F_mode(:,:), beam_mat(:,:)
  real(kind=8), allocatable :: beam_mat_bin(:,:),beam_mode(:,:),beam_eval(:)
  real(kind=8), allocatable :: a(:), b(:,:), c(:)
  logical :: beam_include_beam_modes = .true.
contains
  
  subroutine init_act_beam_chisq(win_func,bval,blmin,blmax)
    
    use act_options
    implicit none
    real(kind=8), dimension(0:bmax0-1), intent(in) :: bval,blmin,blmax
    real(kind=8),dimension(0:bmax0-1,1:tt_lmax), intent(in) :: win_func
    real(8) :: wl(0:bmax0-1), wl_ell(0:tt_lmax)
    integer :: il,i, l, lun
    CHARACTER(LEN=240) :: wfilename
    LOGICAL  :: good
    real :: dummy
    integer :: dummy_int
    
    nmodes = 0
    if (beam_include_beam_modes) nmodes = nmodes + n_beam_modes
    if ( nmodes == 0 ) return
    
    allocate( a(nmodes), b(nmodes,nmodes),c(nmodes) )
    allocate( beam_mat(0:tt_lmax,n_beam_modes),beam_mat_bin(0:bmax0-1,n_beam_modes))
    allocate(beam_mode(0:bmax0-1,n_beam_modes),beam_eval(n_beam_modes) )
    allocate( mode(0:bmax0-1,nmodes),F_mode(0:bmax0-1,nmodes) )
    
    if ( beam_include_beam_modes ) then

       wfilename  = trim(ACT_data_dir)//'Data/act_window_148x148_v2a1s.dat'
    
       inquire(file=wfilename,exist = good)
       if(.not.good)then
          write(*,*) 'cant find', trim(wfilename), trim(ACT_data_dir)
          stop
       endif
       call get_free_lun( lun )
       open(unit=lun,file=wfilename,form='formatted',status='unknown',action='read')
       
       do i=0,tt_lmax
          read(lun,*) dummy_int, wl_ell(i)
       enddo
       close(lun)
       !normalize back to l=2 to match the window used in spectrum 
       wl_ell(0:tt_lmax)=wl_ell(0:tt_lmax)/wl_ell(2) 

       !Multiply window function by bandpower window matrix
       wl(0:bmax0-1)=matmul(win_func(0:bmax0-1,1:tt_lmax),wl_ell(1:tt_lmax))
     
       call get_free_lun( lun )
       open(unit=lun,file=trim(ACT_data_dir)//'Data/top_10_evals_148x148_v2a1s.dat',form='formatted',status='unknown',action='read')
       read(lun,*),beam_eval(1:n_beam_modes)
       close(lun)
            
       call get_free_lun( lun )
       open(unit=lun,file=trim(ACT_data_dir)//'Data/top_10_modes_148x148_v2a1s.dat',form='formatted',status='unknown',action='read')

       do i = 1,n_beam_modes
          do il =0,tt_lmax
             read(lun,*) beam_mat(il,i)
          end do
          beam_mat_bin(0:bmax0-1,i)=matmul(win_func(0:bmax0-1,1:tt_lmax),beam_mat(1:tt_lmax,i))
       enddo
       close(lun)

       do i=1, n_beam_modes
             beam_mode(0:bmax0-1,i)=beam_mat_bin(0:bmax0-1,i)*sqrt(beam_eval(i))/sqrt(wl(0:bmax0-1))*2.
       enddo
       
    end if
    
  end subroutine init_act_beam_chisq

  function compute_act_beam_chisq(bmin1,bmax1, btt_th, btt_dat, fisher)

    USE ACT_options
    implicit none
    
    integer, intent(in) :: bmin1, bmax1
    real(kind=8), dimension(0:bmax0-1), intent(in) :: btt_th, btt_dat
    real(kind=8), dimension(0:bmax0-1,0:bmax0-1), intent(in) :: fisher
    real(kind=8) :: compute_act_beam_chisq
    real(kind=8) :: dgauss,  dlndet
    integer :: i, j, l, l1, l2, stat,il

    if ( nmodes == 0 ) then
       compute_act_beam_chisq = 0d0
       return
    end if
    
    mode = 0d0
    if ( beam_include_beam_modes ) then
       do i = 1,n_beam_modes
          do l = bmin1,bmax1
             mode(l,i) = beam_mode(l,i)*btt_th(l)
          end do
       end do
    end if
   
    F_mode = 0d0
    do i = 1,nmodes
       do l = bmin1,bmax1
          F_mode(l,i) = fisher(l,l)*mode(l,i)
       end do
    end do
    
    a = 0d0
    b = 0d0
    do i = 1,nmodes
       do l = bmin1,bmax1
          a(i) = a(i) + (btt_dat(l)-btt_th(l))*F_mode(l,i)
       end do
       
       b(i,i) = b(i,i) + 1d0
       do j = i,nmodes
          do l = bmin1,bmax1
             b(i,j) = b(i,j) + mode(l,i)*F_mode(l,j)
          end do
          if ( i /= j ) b(j,i) = b(i,j)
       end do
    end do
    call dpotrf( 'L', nmodes, b, nmodes, stat )
    c(:) = a(:)
    call dpotrs( 'L', nmodes, 1, b, nmodes, c, nmodes, stat )
    
    dgauss = 0d0
    dlndet =1.d0
    do i = 1,nmodes
       dgauss = dgauss + a(i)*c(i)
       dlndet = dlndet*b(i,i)**2
    end do

    dlndet = dlog(dlndet)
    compute_act_beam_chisq = -dgauss + dlndet
    
  end function compute_act_beam_chisq

  subroutine get_free_lun( lun )

    implicit none
    integer, intent(out) :: lun
    
    integer, save :: last_lun = 19
    logical :: used
    lun = last_lun
    do
       inquire( unit=lun, opened=used )
       if ( .not. used ) exit
       lun = lun + 1
    end do
    
    last_lun = lun
  end subroutine get_free_lun

end module act_beam_chisq

